/*
 * chains.c
 * This module contains stuff that ought to get macroized.
 * [eichin:19880331.2032EST]
 */

#macro EXPAND(refptr)
{
  ASSERT(refptr->tag.CTL == END);
  if (refptr->data.CTL == end)
    {
      ropeptr tmpptr;
      
      refptr->tag.CTL = NEXT_FORWARD;

      tmpptr = getblock();
      ASSERT(refptr->tag.CTL == NEXT_BACKWARD);
      tmpptr->data.NEXT = refptr-1;

      refptr = refptr->data.NEXT = tmpptr+1;
    }
}
#endmacro

#macro INC(refptr)
{
  if(((++refptr)->tag.CTL == NEXT_FORWARD)
     || (refptr->tag.CTL == NEXT_FORWARD_UNALIGNED))
    {
      refptr = refptr->data.NEXT;
    }
}
#endmacro
#macro START(refptr) INC(refptr)
#endmacro
/*
 * DEC: assumes that the refptr is NOT at the first location.
 */
#macro DEC(refptr)
{
  if(((--refptr)->tag.CTL == NEXT_BACKWARD)
     || (refptr->tag.CTL == NEXT_BACKWARD_UNALIGNED))
    {
      refptr = refptr->data.NEXT;
    }
}
#endmacro
/*
 * no assumptions?
 */
#macro insert_cell(refptr)
{
  rope_elem cpy, cpy2;
  cpy.tag.CTL = END;
  cpy.data.CTL = end;

  while(refptr->tag.CTL != END)
    {
      cpy2 = *refptr;
      *refptr = cpy;
      cpy = cpy2;
      INC(refptr);
    }
  EXPAND(refptr);
  *refptr = cpy;
}
#endmacro

/*
 * Here some things get a tad confusing:
 * NEXT_FORWARD: the next block is begun at the beginning, this is at
 * 		the end.
 * NEXT_BACKWARD: the previous block is at the end, this is
 * 		at the beginning.
 * NEXT_FORWARD_UNALIGNED: the next block is begun at the beginning,
 * 		this is NOT at end.
 * NEXT_BACKWARD_UNALIGNED: this is at the beginning, but the previous
 * 		block is NOT at end.
 *
 * So, blocks always begin at the beginning, but they may end at
 * different places; we must always be able to tell from either end
 * what to do. [though perhaps we could just know at one end...]
 */

#macro compress(refptr)
{
  ASSERT(refptr->tag.CTL == END);
  refptr--;
  if (refptr->tag.CTL == NEXT_BACKWARD)
    {
      ropeptr savptr;
      savptr = refptr->data.NEXT;
      make_free(refptr);
      refptr = savptr++;	/* savptr is the tail of the BACK block */
      savptr->tag.CTL = END;
      savptr->data.CTL = end;
    }
  if (refptr->tag.CTL == NEXT_BACKWARD_UNALIGNED)
    {
      ropeptr savptr;
      savptr = refptr->data.NEXT;
      refptr->tag.CTL = END;
      refptr->data.CTL = end;
      refptr = savptr++;	/* savptr is NOT the tail...*/
      savptr->tag.CTL = END;
      savptr->data.CTL = notend;
    }
}
#endmacro
/*
 * useup: takes a pointer (to a `beginning' of a list) and eats the
 * current element; if it skips a block, it frees that. Maintains the
 * `BEGIN' pointer as needed.
 *
 * It should really check the following conditions:
 * 	NEXT_FORWARD_UNALIGN (and free on it)
 * 	END.notend in the new item: free the block AND `break'...
 * 	END.end in the new item: same?
 */
#macro useup(refptr, ENDHANDLER)
{
  refptr->tag.CTL = END;
  refptr->data.CTL = notend;
  refptr++;
  if (refptr->tag.CTL == END)
    {
      make_free(refptr);
      ENDHANDLER;		/* usually break or continue... */
    }
  if (refptr->tag.CTL == NEXT_FORWARD)
    {
      ropeptr nxtptr;
      nxtptr = refptr->data.NEXT;
      /*       s2ptr = refptr;??? */
      refptr->tag.CTL = END;
      refptr->data.CTL = end;
      while(refptr->tag.CTL != BEGIN)
	{
	  refptr --;
	}
      (--nxtptr)->tag.CTL = BEGIN;
      nxtptr->data.NEXT = refptr->data.NEXT;
      make_free(refptr);

      refptr = nxtptr+1;
    }
}
#endmacro
/*
 * back_trace: goes back to the start of this `block' and then frees
 * the whole thing. Assumes a BEGIN tag, so that it can actually free
 * it...
 */
#macro back_trace(refptr)
{
  while(refptr->tag.CTL != BEGIN)
    {
      refptr --;
    }
  make_free(refptr);
}
#endmacro

#declare
extern ropeptr nextfree;
#enddeclare

#macro make_free(refptr)
{
  refptr->tag.CTL = FREELIST;
  refptr->data.NEXT = nextfree;
  nextfree = refptr;
}
#endmacro

/*
 * Since we spec rope[0] = { BEGIN, endptr }, anything that adds to
 * the end of a chain just needs to reference through the first
 * element. Such as:
 * wire->data.NEXT->{data,tag}
 */

#macro set_item(wire, TYPE, piece, partname, value)
{
  EXPAND(wire);
  wire->tag.TYPE = piece;
  (wire->data.TYPE.partname) = value;
}
#endmacro
/*
 * special case for events:
 */
#macro set_event(evptr, value, time)
{
  EXPAND(evptr);
  evptr->tag.EVENT = value;
  evptr->data.EVENT = time;
}
#endmacro

#macro add_item(wire, TYPE, piece, partname, value)
{
  set_item(wire->data.NEXT, TYPE, piece, partname, value);
  INC(wire->data.NEXT);
}
#endmacro

#macro add_event(timlin, value, time)
{
  set_event(timlin->data.NEXT, value, time);
  INC(timlin->data.NEXT);
}
#endmacro

#macro add_chain(wire, TYPE, piece, partname)
{
  add_item(wire, TYPE, piece, partname, getblock());
}
#endmacro

#macro find_item(wire, TYPE, value)
{
  while(wire->tag.TYPE != END)
    {
      if(wire->tag.TYPE == value) break;
      INC(wire);
    }
}
#endmacro

#macro nextwire(argv, wire, time)
{
  wire = argv->data.GENERIC.genwire;
  ASSERT(argv->tag.GENERIC == genwire);
  find_item(wire, WIRE, FUTURE);
  time = wire->data.WIRE.future;
  INC(argv);
}
#endmacro

#macro readnextwire(argv, wire, time)
{
  wire = argv->data.GENERIC.genwire;
  ASSERT(argv->tag.GENERIC == genwire);
  find_item(wire, WIRE, NOW);
  time = wire->data.WIRE.now;
  INC(argv);
}
#endmacro
